import numpy as np
from .Image import Image


def detectpeak(
    img: Image, rms: float, threshold_rms: int = 5, find_max: bool = True
) -> list[tuple[int, int, float]]:
    """
    Detects peaks in an image.

    Args:
        img (Image): The image object.
        rms (float): The RMS noise level for peak detection. Only pixels with intensity greater than threshold_rms * rms will be considered.
        threshold_rms (int): The threshold in terms of RMS to consider a pixel as a peak. Default is 5.
        find_max (bool): If True, the function will find the maximum peaks. Otherwise, it will find the minimum peaks.

    Returns:
        list[tuple[int, int, float]]: The list of detected peaks. Each peak is represented as a tuple (x, y, value).
    """
    # beam and search cell size
    img.convert_axes_unit("arcsec")
    if img.beam is None:
        raise ValueError("The image does not have a beam size.")
    if img.incr_x is None or img.incr_y is None:
        raise ValueError("Image increment x or y is None.")
    beam_x = img.beam[0] / abs(img.incr_x)
    beam_y = img.beam[1] / abs(img.incr_y)
    beam_ang = (90 + img.beam[2]) * np.pi / 180
    cell_width = int(
        np.sqrt((beam_x * np.cos(beam_ang)) ** 2 + (beam_x * np.sin(beam_ang)) ** 2) + 1
    )
    cell_height = int(
        np.sqrt((beam_x * np.sin(beam_ang)) ** 2 + (beam_y * np.cos(beam_ang)) ** 2) + 1
    )

    # detect peaks
    peak: list[tuple[int, int, float]] = []

    if img.height is None or img.width is None:
        raise ValueError("Image height or width is None.")
    if img.data is None:
        raise ValueError("Image data is None.")

    threshold = threshold_rms * rms
    data = img.get_two_dim_data()

    for i in range(cell_height // 2, img.height - cell_height // 2, 1):
        for j in range(cell_width // 2, img.width - cell_width // 2, 1):
            if data[i][j] > threshold:
                region = data[
                    i - cell_height // 2 : i + cell_height // 2 + 1,
                    j - cell_width // 2 : j + cell_width // 2 + 1,
                ]
                if find_max:
                    if data[i][j] == np.max(region):
                        peak.append((j, i, data[i][j]))
                else:
                    if data[i][j] == np.min(region):
                        peak.append((j, i, data[i][j]))

    return peak



import numpy as np
from .utilities import unitConvDict


class Image:
    def __init__(self):
        self.imagename: str | None = None
        self.data: np.ndarray | None = None
        self.width: int | None = None
        self.height: int | None = None
        self.nchan: int | None = None
        self.center_radec: tuple[float, float] | None = None  # (RA, Dec)
        self.center_pix: tuple[float, float] | None = None  # (X, Y)
        self.freq0: float | None = None
        self.incr_x: float | None = None
        self.incr_y: float | None = None
        self.incr_hz: float | None = None
        self.unit_x: str | None = None
        self.unit_y: str | None = None
        self.unit_data: str | None = None
        self.beam: tuple[float, float, float] | None = None  # (major, minor, angle)

    def convert_axes_unit(self, unit: str):
        """
        Convert the axes units of the image to the specified unit.

        Args:
            unit (str): The unit to convert to (e.g., 'arcsec').
        """
        try:
            if self.unit_x is not None:
                self.incr_x *= unitConvDict[(self.unit_x, unit)]
            else:
                raise ValueError("Unit of x-axis is None.")
        except KeyError:
            raise ValueError(
                f"Unsupported unit conversion from {self.unit_x} to {unit}."
            )
        try:
            if self.unit_y is not None:
                self.incr_y *= unitConvDict[(self.unit_y, unit)]
            else:
                raise ValueError("Unit of y-axis is None.")
        except KeyError:
            raise ValueError(
                f"Unsupported unit conversion from {self.unit_y} to {unit}."
            )
        self.unit_x = unit
        self.unit_y = unit

    def get_ticks(
        self, xtickspan: int, ytickspan: int, relative: bool, fmt: str, width: int | None = None, height: int | None = None
    ) -> tuple[list[float], list[str], list[float], list[str]]:
        """
        Returns x and y ticks and tick labels.

        Args:
            xtickspan (int): Span of ticks of x-axis.
            ytickspan (int): Span of ticks of y-axis.
            relative (bool): If True, ticks are relative coordinates. If False, ticks are global coordinates (NOT IMPLEMENTED YET!).
            fmt (str): Format of tick labels.
            width (int | None): Width of the image. If None, use the width of the image.
            height (int | None): Height of the image. If None, use the height of the image.

        Returns:
            xticks, xticks_label, yticks, yticks_label
        """
        _fmt = "{" + fmt + "}"
        if self.width is None or self.height is None:
            raise ValueError("Image width and height is None.")
        if self.incr_x is None or self.incr_y is None:
            raise ValueError("Image increment x and y is None.")
        if width is None:
            width = self.width
        if height is None:
            height = self.height
        xini = (self.width - width) / 2
        yini = (self.height - height) / 2
        xmid = self.width / 2
        ymid = self.height / 2
        xfin = (self.width + width) / 2
        yfin = (self.height + height) / 2
        xlini = -width / 2 * self.incr_x
        ylini = -height / 2 * self.incr_y
        xlmid = 0
        ylmid = 0
        xlfin = width / 2 * self.incr_x
        ylfin = height / 2 * self.incr_y
        if relative:
            xticks_label_num: list[float] = [xlini, 0.0]
            yticks_label_num: list[float] = [ylini, 0.0]
        else:
            raise NotImplementedError(
                "Global coordinates are not implemented yet. Use relative coordinates."
            )
        xticks = [xini, xmid]
        yticks = [yini, ymid]
        for i in range(1, xtickspan + 1):
            xticks.append((xmid - xini) * i / (xtickspan + 1) + xini)
            xticks_label_num.append((xlmid - xlini) * i / (xtickspan + 1) + xlini)
            xticks.append((xmid - xfin) * i / (xtickspan + 1) + xfin)
            xticks_label_num.append((xlmid - xlfin) * i / (xtickspan + 1) + xlfin)
        for i in range(1, ytickspan + 1):
            yticks.append((ymid - yini) * i / (ytickspan + 1) + yini)
            yticks_label_num.append((ylmid - ylini) * i / (ytickspan + 1) + ylini)
            yticks.append((ymid - yfin) * i / (ytickspan + 1) + yfin)
            yticks_label_num.append((ylmid - ylfin) * i / (ytickspan + 1) + ylfin)
        xticks_label: list[str] = [""] * len(xticks)
        yticks_label: list[str] = [""] * len(yticks)
        for i, s in enumerate(xticks_label_num):
            xticks_label[i] = _fmt.format(s)
        for i, s in enumerate(yticks_label_num):
            yticks_label[i] = _fmt.format(s)
        return xticks, xticks_label, yticks, yticks_label

    def get_two_dim_data(self, stokes: int = 0, chan: int = 0) -> np.ndarray:
        """
        Extracts the 2D data based on the specified Stokes and channel indices.

        Args:
            stokes (int, optional): Stokes parameter index. Defaults to 0.
            chan (int, optional): Channel index. Defaults to 0.

        Returns:
            np.ndarray: The extracted 2D data.
        """
        if self.data is None:
            raise ValueError("Image data is None.")
        if stokes < 0 or stokes >= self.data.shape[0]:
            raise IndexError(
                f"Stokes index {stokes} is out of bounds for the image data."
            )
        if chan < 0 or chan >= self.data.shape[1]:
            raise IndexError(
                f"Channel index {chan} is out of bounds for the image data."
            )

        if self.data.ndim == 4:
            return self.data[stokes, chan]
        elif self.data.ndim == 3:
            return self.data[chan]
        elif self.data.ndim == 2:
            return self.data
        else:
            raise ValueError("Unsupported image data dimensions.")

    def keep_stokes_chan(self, stokes: int, chan: int):
        """
        Keep only specific stokes and channel from the image data.
        This is useful for the image that has only one stokes and channel.

        Args:
            stokes (int): The Stokes parameter to keep.
            chan (int): The channel to keep.
        """
        if self.data is None:
            raise ValueError("Image data is None.")
        # Check if the data has four dimensions (Stokes, Channel, Y, X)
        if not self.data.ndim == 4:
            raise ValueError(
                f"Data must be 4D (Stokes, Channel, Y, X), but got {self.data.ndim}D."
            )
        # Check if the specified stokes and channel are valid
        if stokes < 0 or chan < 0:
            raise ValueError("Stokes and channel indices must be non-negative.")
        if stokes >= self.data.shape[0]:
            raise ValueError(
                f"Stokes index {stokes} is out of bounds for data with shape {self.data.shape}."
            )
        if chan >= self.data.shape[1]:
            raise ValueError(
                f"Channel index {chan} is out of bounds for data with shape {self.data.shape}."
            )
        if self.data.ndim == 4:
            self.data = self.data[stokes, chan]
        elif self.data.ndim == 3:
            self.data = self.data[chan]
        else:
            raise ValueError(
                f"Unsupported data dimension: {self.data.ndim}. Expected 3D or 4D data."
            )


import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from astropy.coordinates import SkyCoord
from astropy import units as u
from .Image import Image


def load_fits(
    fits_file: str,
    width: int | None = None,
    height: int | None = None,
    center_radec: tuple[float, float] | None = None,
) -> Image:
    """
    Create an Image object from a FITS file.

    Args:
        fits_file (str): Path to the FITS file.
        width (int, optional): Width of the cropped image. If None, uses the full width.
        height (int, optional): Height of the cropped image. If None, uses the full height.
        center_radec (tuple[float, float], optional): Center coordinates in RA, Dec format. If None, uses the center of the image.

    Returns:
        Image: An Image object.
    """
    image = Image()

    # remove the extension from the filename
    image.imagename = fits_file.rsplit(".", 1)[0]
    # Load the FITS file and extract the data and header information
    try:
        with fits.open(fits_file) as hdul:
            image.data = np.array(hdul[0].data)  # type: ignore
            header = hdul[0].header  # type: ignore
            image.width = header.get("NAXIS1", image.data.shape[1])
            image.height = header.get("NAXIS2", image.data.shape[0])
            image.nchan = header.get("NAXIS3", 1)
            image.center_radec = (header.get("CRVAL1", 0.0), header.get("CRVAL2", 0.0))
            image.center_pix = (header.get("CRPIX1", image.width // 2) - 1, header.get("CRPIX2", image.height // 2) - 1)
            image.freq0 = header.get("CRVAL3", 0.0)
            image.incr_x = header.get("CDELT1", 0.0)
            image.incr_y = header.get("CDELT2", 0.0)
            image.incr_hz = header.get("CDELT3", 0.0)
            image.unit_x = header.get("CUNIT1", "deg")
            image.unit_y = header.get("CUNIT2", "deg")
            image.unit_data = header.get("BUNIT", "Jy/beam")
            beam_major = (
                header.get("BMAJ", 0.0) * 3600
            )  # Convert from degrees to arcseconds
            beam_minor = (
                header.get("BMIN", 0.0) * 3600
            )  # Convert from degrees to arcseconds
            beam_angle = header.get("BPA", 0.0)
            if beam_major and beam_minor:
                image.beam = (beam_major, beam_minor, beam_angle)
            else:
                image.beam = None

            # Cropping
            if center_radec is not None:
                wcs = WCS(header)
                center_coord = SkyCoord(
                    center_radec[0],
                    center_radec[1],
                    unit=(u.hourangle, u.deg),
                    frame="icrs",
                )
                image.center_pix = wcs.world_to_pixel(center_coord)
                # Debug
                # print(
                #     f"New center pixel coordinates: {image.center_pix[0]}, {image.center_pix[1]}"
                # )
            # new width and height
            if width is None:
                width = image.width
            if height is None:
                height = image.height
            # Calculate new left, right, top, bottom based on the new center
            # print(image.center_pix, width, height)
            left = int(round(image.center_pix[0] - width / 2))
            right = int(round(image.center_pix[0] + width / 2))
            bottom = int(round(image.center_pix[1] - height / 2))
            top = int(round(image.center_pix[1] + height / 2))
            # Ensure the new dimensions are within bounds
            if left < 0:
                right += left
                left = 0
            if right > image.width:
                left += right - image.width
                right = image.width
            if bottom < 0:
                top += bottom
                bottom = 0
            if top > image.height:
                bottom += top - image.height
                top = image.height
            # Debugging output
            # print(
            #     f"Cropping to: left={left}, right={right}, bottom={bottom}, top={top}"
            # )
            # Crop the data
            image.data = image.data[bottom:top, left:right]

            # For debugging purposes, print all attributes
            # print(f"Loaded FITS file '{fits_file}':")
            # print(f"  Width: {image.width}, Height: {image.height}, Channels: {image.nchan}")
            # print(f"  Center (RA, Dec): {image.center_radec}")
            # print(f"  Center (X, Y): {image.center_pix}")
            # print(f"  Freq0: {image.freq0}")
            # print(f"  Incr X: {image.incr_x}, Incr Y: {image.incr_y}, Incr Hz: {image.incr_hz}")
            # print(f"  Units: X={image.unit_x}, Y={image.unit_y}, Data={image.unit_data}")
            # if image.beam:
            #     print(f"  Beam: Major={image.beam[0]}, Minor={image.beam[1]}, Angle={image.beam[2]}")
            # else:
            #     print("  No beam information available.")
    except Exception as e:
        raise ValueError(f"Failed to open FITS file '{fits_file}': {e}")

    return image


def load_image(
    imagename: str,
    width: int = None,
    height: int = None,
    center_radec: tuple[float, float] = None,
) -> Image:
    """
    Create an Image object from a CASA image file.

    Args:
        imagename (str): Path to the CASA image file.
        width (int, optional): Width of the cropped image. If None, uses the full width.
        height (int, optional): Height of the cropped image. If None, uses the full height.
        center_radec (tuple[float, float], optional): Center coordinates in RA, Dec format. If None, uses the center of the image.

    Returns:
        Image: An Image object.
    """
    from casatasks import imhead
    from casatools import image as casa_image, measures

    image = Image()

    # info (imhead)
    image.imagename = imagename
    inform = imhead(imagename=imagename)
    if inform is None or not isinstance(inform, dict):
        raise ValueError(f'Failed to retrieve image inform from "{imagename}".')

    image.width = inform["shape"][0]
    image.height = inform["shape"][1]
    image.nchan = inform["shape"][3]

    if image.width is None or image.height is None:
        raise ValueError(f'Invalid image shape for "{imagename}".')

    image.x0 = inform["refval"][0]
    image.y0 = inform["refval"][1]
    image.freq0 = inform["refval"][3]

    image.incr_x = inform["incr"][0]
    image.incr_y = inform["incr"][1]
    image.incr_hz = inform["incr"][3]
    try:
        if inform["restoringbeam"]["positionangle"]["unit"] == "rad":
            pa = np.rad2deg(
                inform["restoringbeam"]["positionangle"]["value"]
            )
        else:
            pa = inform["restoringbeam"]["positionangle"]["value"]
        image.beam = (
            inform["restoringbeam"]["major"]["value"],
            inform["restoringbeam"]["minor"]["value"],
            pa,
        )
    except KeyError:
        image.beam = None
    image.unit_data = inform["unit"]
    image.unit_x = inform["axisunits"][0]
    image.unit_y = inform["axisunits"][1]

    # figure size
    if width is None or width <= 0 or width > image.width:
        width = image.width
    if (
        height is None
        or height <= 0
        or height > image.height
    ):
        height = image.height

    # Load data
    ia = casa_image()
    ia.open(imagename)
    if center_radec is not None:
        me = measures()
        center_coord = me.direction("J2000", v0=center_radec[0], v1=center_radec[1])
        pix_center = ia.topixel(
            [center_coord["m0"]["value"], center_coord["m1"]["value"]]
        )
        x_center, y_center = int(pix_center["numeric"][0]), int(
            pix_center["numeric"][1]
        )
    else:
        x_center, y_center = image.width // 2, image.height // 2
    image.center_pix = (x_center, y_center)
    blc_x, blc_y = x_center - width // 2, y_center - height // 2
    trc_x, trc_y = x_center + width // 2, y_center + height // 2
    d_blc_x, d_blc_y = -blc_x, -blc_y
    d_trc_x, d_trc_y = trc_x - image.width, trc_y - image.height
    if d_blc_x > 0:
        blc_x = 0
        trc_x -= d_blc_x
    if d_trc_x > 0:
        trc_x = image.width
        blc_x += d_trc_x
    if d_blc_y > 0:
        blc_y = 0
        trc_y -= d_blc_y
    if d_trc_y > 0:
        trc_y = image.height
        blc_y += d_trc_y
    blc = [blc_x, blc_y]
    trc = [trc_x, trc_y]
    image.height = trc_y - blc_y
    image.width = trc_x - blc_x
    rawdata = ia.getchunk(blc=blc, trc=trc)
    ia.close()

    # Need to reconsider the shape of image.img
    rawdata = rawdata.transpose(2, 3, 1, 0)
    image.data = rawdata

    # Convert RA/DEC units to arcsec
    image.convert_axes_unit("arcsec")

    return image


from math import degrees, radians
import numpy as np
from .Image import Image

def azimuth_cut(
        img: Image,
        radius: float,
        inclination: float,
        pa: float,
        beam_factor: float = 0.5,
        stokes: int = 0,
        chan: int = 0
) -> tuple:
    """
    Extract an azimuthal cut from an image using a deprojected elliptical annulus.

    This implementation computes the deprojected radius for each pixel,
    given the inclination and position angle, and selects pixels that lie
    within an annulus defined by [radius_px - beam_size/2, radius_px + beam_size/2].
    The azimuthal angle is computed in the deprojected frame.

    Args:
        img (Image): The Image object.
        radius (float): The radius in arcsec.
        inclination (float): The inclination angle in degrees.
        pa (float): The position angle in degrees.
        beam_factor (float, optional): Sampling size based on the beam size. Defaults to 0.5.
        stokes (int, optional): Stokes parameter index. Defaults to 0.
        chan (int, optional): Channel index. Defaults to 0.

    Returns:
        tuple: A tuple of three numpy arrays:
            - The azimuthal angle (in degrees) bins.
            - The mean intensity in each bin.
            - The standard deviation of the intensity in each bin.
    """
    # Calculate the center of the image
    # center_x = img.width // 2
    # center_y = img.height // 2

    # Calculate the beam size in pixel units (assuming beam and increments are defined)
    if not img.beam:
        raise ValueError("The image does not have a beam size.")
    img.convert_axes_unit('arcsec')
    if img.incr_x is None or img.incr_y is None:
        raise ValueError("Image increments are not defined.")
    beam_x = img.beam[0] / np.abs(img.incr_x)
    beam_y = img.beam[1] / np.abs(img.incr_y)
    beam_size = max(beam_x, beam_y)

    # Convert the input radius from arcsec to pixels.
    radius_px = radius / np.abs(img.incr_x)

    # Define the radial range (annulus width) in deprojected pixel units.
    r_min = radius_px - beam_size / 2
    r_max = radius_px + beam_size / 2

    # Sampling step in the azimuthal direction.
    sampling_rad = beam_size * beam_factor / radius_px
    sampling_deg = degrees(sampling_rad)
    line_azm = np.arange(0, 360, sampling_deg)
    sample = [[] for _ in range(len(line_azm))]

    # Convert inclination and PA to radians
    incl_rad = radians(inclination)
    pa_rad = radians(pa)

    # Loop over image pixels
    # For each pixel, compute its deprojected radius and azimuth angle.
    if img.data is None:
        raise ValueError("Image data is not loaded.")
    data = img.get_two_dim_data(stokes=stokes, chan=chan)
    if img.height is None or img.width is None:
        raise ValueError("Image dimensions are not defined.")
    if img.center_pix is None:
        img.center_pix = (img.width // 2, img.height // 2)
    for i in range(img.height):
        for j in range(img.width):
            dx = j - img.center_pix[0]
            dy = i - img.center_pix[1]

            # Rotate coordinates by PA to align with the disk's axes.
            xp = dx * np.cos(pa_rad) + dy * np.sin(pa_rad)
            yp = -dx * np.sin(pa_rad) + dy * np.cos(pa_rad)

            # Compute deprojected radius:
            # y'-coordinate is corrected by 1/cos(inclination)
            r_deproj = np.sqrt(xp**2 + (yp / np.cos(incl_rad))**2)

            # Select pixels within the annulus.
            if r_deproj < r_min or r_deproj > r_max:
                continue

            # Compute the deprojected azimuth angle (in degrees).
            # Note: np.arctan2(y, x) yields the angle in radians.
            theta_deproj = (np.degrees(np.arctan2(yp / np.cos(incl_rad), xp)) - 90) % 360

            # Find the closest azimuthal bin.
            idx = np.argmin(np.abs(line_azm - theta_deproj))
            sample[idx].append(data[i, j])

    # Compute mean and standard deviation in each azimuthal bin.
    line_mean = np.array([np.mean(s) if s else np.nan for s in sample])
    line_std = np.array([np.std(s) if s else np.nan for s in sample])

    return line_azm, line_mean, line_std


import numpy as np
from .Image import Image


def imstat(
    image: Image,
    uncertainty: float,
    unit: str = "beam",
    mask: np.ndarray | None = None,
    inverse_mask: bool = False,
):
    """
    Alternative version of imstat.

    Args:
        image (Image): Image object.
        uncertainty (float): Flux uncertainty.
        unit (str, optional): Unit of area. Choices are 'beam' or 'arcsec'. Defaults to 'beam'.
        mask (np.ndarray, optional): If specified, statistics are calculated using only the data within the mask. Defaults to None.
        inverse_mask (bool, optional): If True, the specified mask region is inverted. Defaults to False.
    """
    data = image.data
    image.convert_axes_unit("arcsec")
    if unit == "beam":
        unit_area = 1
    elif unit == "arcsec":
        unit_area = np.pi * image.beam[0] * image.beam[1]
    else:
        raise ValueError("Arg `unit` must be `'beam'` or `'arcsec'`.")
    data /= unit_area
    ret = {}
    ret["unit"] = image.unit_data
    ret["restoring beam"] = {}
    ret["restoring beam"]["x"] = image.beam[0]
    ret["restoring beam"]["y"] = image.beam[1]
    ret["restoring beam"]["ang"] = image.beam[2]
    ret["all"] = {}
    ret["all"]["max"] = np.max(data)
    ret["all"]["max_sigma"] = ret["all"]["max"] * uncertainty
    ret["all"]["min"] = np.min(data)
    ret["all"]["sum"] = np.sum(data)
    ret["all"]["sumsq"] = np.sum(np.square(data))
    ret["all"]["mean"] = np.mean(data)
    ret["all"]["sigma"] = np.var(data, ddof=1)
    ret["all"]["rms"] = np.sqrt(ret["all"]["sumsq"] / data.size)
    # process mask
    if mask is not None:
        # Normalize the mask to True/False
        if mask.dtype != bool:
            mask = mask.astype(bool)
        if inverse_mask:
            mask = ~mask
        # Mask the data as np.nan
        data = np.where(mask, data, np.nan)
        # count the number of unmasked pixels
        num_unmasked = np.sum(mask)
        ret["masked"] = {}
        ret["masked"]["max"] = np.nanmax(data)
        ret["masked"]["min"] = np.nanmin(data)
        ret["masked"]["sum"] = np.nansum(data)
        ret["masked"]["sumsq"] = np.nansum(np.square(data))
        ret["masked"]["mean"] = np.nanmean(data)
        ret["masked"]["sigma"] = np.nanvar(data, ddof=1)
        ret["masked"]["rms"] = np.sqrt(ret["masked"]["sumsq"] / num_unmasked)
        ret["psnr"] = ret["all"]["max"] / ret["masked"]["rms"]
        ret["psnr_sigma"] = ret["all"]["max_sigma"] / ret["masked"]["rms"]
    return ret



import matplotlib
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from .utilities import get_si_prefix_base10, get_si_prefix_symbol


def set_cbar(im: matplotlib.image.AxesImage, observable: str, unit: str, rescale: str, fmt: str, fmt_default: str) -> None:
    """
    Sets the colorbar for the given figure and image axes.

    Args:
        im (plt.image.AxesImage): The image axes.
        observable (str): The observable being plotted.
        unit (str): The unit of the observable.
        rescale (str): The SI prefix to rescale the colorbar.
        fmt (str): The format string for the colorbar labels.
        fmt_default (str): The default format string for the colorbar labels.
    """
    def fn_fmt(x, pos):
        try:
            return ('{' + fmt + '}').format(x * get_si_prefix_base10(rescale))
        except ValueError:
            return ('{' + fmt_default + '}').format(x * get_si_prefix_base10(rescale))
    cbar = plt.colorbar(im, format=ticker.FuncFormatter(fn_fmt))
    if observable is None:
        label_observable = 'Intensity'
    else:
        label_observable = observable
    if unit is None or unit == '':
        label_unit = ''
    else:
        label_unit = f'[{get_si_prefix_symbol(rescale) + unit}]'
    cbar.set_label(f'{label_observable} {label_unit}')


def set_axes_options(ax: matplotlib.axes.Axes, title: str, xlabel: str, ylabel: str, xticks, xticks_label, yticks, yticks_label):
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticks_label)
    ax.set_yticks(yticks)
    ax.set_yticklabels(yticks_label)



import numpy as np
from matplotlib.patches import Ellipse
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from scipy.ndimage import zoom
from .Image import Image

# from .PlotConfig import PlotConfig
from .utilities import unitDict  # , get_pret_dir_name
from .matplotlib_helper import set_cbar, set_axes_options

# from .prepare_image import prepare_image


def draw_beam(
    ax: Axes, img: Image, beam: tuple | None = None, color: str = "white"
) -> None:
    """
    Draws the beam on the image.

    Args:
        ax (matplotlib.axes.Axes): The Axes object.
        img (Image): The Image object.
        beam (tuple): The beam parameters (beam_x, beam_y, beam_ang).
                      If both img and beam are provided, beam will be used.
        color (str): The color of the beam. Default to 'white'.
    """
    x_min, x_max = ax.get_xlim()
    y_min, y_max = ax.get_ylim()
    beam_pos_x = x_min + (x_max - x_min) / 8
    beam_pos_y = y_min + (y_max - y_min) / 8
    if img.beam is not None and img.incr_x is not None and img.incr_y is not None:
        beam_x = img.beam[0] / np.abs(img.incr_x)
        beam_y = img.beam[1] / np.abs(img.incr_y)
        beam_ang = img.beam[2]
    elif beam is not None:
        beam_x, beam_y, beam_ang = beam
    else:
        print("This image does not have a beam.")
        return
    ellipse = Ellipse(
        xy=(beam_pos_x, beam_pos_y),
        width=beam_x,
        height=beam_y,
        angle=90 + beam_ang,
        facecolor=color,
        edgecolor=color,
    )
    ax.add_patch(ellipse)


def imshow(ax: Axes, img: Image, **kwargs):
    """
    Rasterizes an image on given Axes object with matplotlib from a CASA style image file.

    Args:
        ax (matplotlib.axes.Axes): The Axes object.
        img (Image): The Image object.
        config (PlotConfig): The PlotConfig object.

    Returns:
        matplotlib.image.AxesImage: The AxesImage object.
    """
    # if config is None:
    #     config = PlotConfig()
    # imagename, img, config = prepare_image(imagename, **kwargs)
    if axisunit := kwargs.get("axisunit"):
        img.convert_axes_unit(axisunit)

    # plot
    if not kwargs.get("show", True):
        plt.ioff()

    # Preparatoin for the cube
    vmax = kwargs.get("vmax")
    vmin = kwargs.get("vmin")

    if img.data is None:
        raise ValueError("Image data is None.")

    if kwargs.get("cbar", "common") == "common":
        if vmin is None:
            vmin = np.nanmin(img.data)
        if vmax is None:
            vmax = np.nanmax(img.data)

    # Specify the stokes and channel
    stokes = kwargs.get("stokes", 0)
    chan = kwargs.get("chan", 0)
    data = img.get_two_dim_data(stokes=stokes, chan=chan)

    im = ax.imshow(
        data,
        cmap=kwargs.get("cmap", "jet"),
        aspect="equal",
        vmin=vmin,
        vmax=vmax,
        origin="lower",
    )
    # Crop the image if specified with width or height
    if width := kwargs.get("width"):
        x_lim = ax.get_xlim()
        x_mid = (x_lim[0] + x_lim[1]) // 2
        ax.set_xlim(x_mid - width // 2, x_mid + width // 2)
    if height := kwargs.get("height"):
        y_lim = ax.get_ylim()
        y_mid = (y_lim[0] + y_lim[1]) // 2
        ax.set_ylim(y_mid - height // 2, y_mid + height // 2)

    # Draw the beam
    draw_beam(ax=ax, img=img, color=kwargs.get("cbeam", "white"))

    # Set the title
    title = kwargs.get("title", "")

    # Set the axes options
    if img.unit_x is None or img.unit_y is None:
        raise ValueError("Image unit_x or unit_y is None.")
    set_axes_options(
        ax,
        title,
        f"RA [{unitDict[img.unit_x]}]",
        f"DEC [{unitDict[img.unit_y]}]",
        *img.get_ticks(
            kwargs.get("xtickspan", 2),
            kwargs.get("ytickspan", 2),
            kwargs.get("relative", True),
            kwargs.get("ticksfmt", ":.3f"),
            width,
            height,
        ),
    )

    # Set the colorbar
    cbarunit = kwargs.get("cbar_unit", img.unit_data)
    set_cbar(
        im,
        kwargs.get("cbar_label", ""),
        cbarunit,
        kwargs.get("rescale", "milli"),
        kwargs.get("cbar_fmt", ":.2f"),
        ":.2f",
    )

    return im


def overlay_contour(
    ax: Axes,
    img_base: Image,
    img: Image,
    stokes: int = 0,
    chan: int = 0,
    fill: bool = False,
    **kwargs,
) -> None:
    """
    Overlays contours on the image.

    To plot with correct scaling, both Image objects should have `incr_x` and `incr_y` attributes.

    Args:
        ax (plt.Axes): The Axes object.
        img_base (Image): Image plotted as background (This should have been already plotted).
        img (Image): Image to be plotted as contours.
        fill (bool): If `True`, the contours will be filled. Default is `False`.
        nchan (int): The channel number to be plotted. If img is a cube, this should be specified.
        **kwargs: Contour configuration keywords of matplotlib.pyplot.contour.
    """
    try:
        im_arr = ax.get_images()[0].get_array()
        if im_arr is None:
            raise ValueError("The axes is not plotted yet.")
        height, width = im_arr.shape
    except IndexError:
        print(
            "The background image is not plotted yet. Please plot the background image first."
        )
        return

    # if img.is_cube and nchan is None:
    #     print('The channel number should be specified for the cube image.')
    #     return

    stokes = kwargs.get("stokes", 0)
    chan = kwargs.get("chan", 0)

    if img_base.unit_x is None:
        raise ValueError("Base image unit_x is None.")
    img.convert_axes_unit(img_base.unit_x)

    if img.data is None:
        raise ValueError("Image data is None.")

    match img.data.ndim:
        case 4:  # Stokes and channel
            data = img.data[stokes, chan]
        case 2:
            data = img.data
        case _:
            raise ValueError(
                f"Unsupported data dimension: {img.data.ndim}. Expected 2D or 4D data."
            )

    if img.incr_x is None or img.incr_y is None:
        raise ValueError("Image increment x or y is None.")
    if img_base.incr_x is None or img_base.incr_y is None:
        raise ValueError("Base image increment x or y is None.")

    ratio = (abs(img.incr_x / img_base.incr_x), abs(img.incr_y / img_base.incr_y))
    data = zoom(data, ratio, order=1)
    # trim the data to the size of the background image by the center
    if data.shape[0] > height:
        diff = data.shape[0] - height
        data = data[diff // 2 : diff // 2 + height]
    if data.shape[1] > width:
        diff = data.shape[1] - width
        data = data[:, diff // 2 : diff // 2 + width]

    # xlim = ax.get_xlim()
    # ylim = ax.get_ylim()
    # extent = ax.get_images()[0].get_extent()

    # x = np.linspace(extent[0], extent[1], data.shape[1])
    # y = np.linspace(extent[2], extent[3], data.shape[0])
    # X, Y = np.meshgrid(x, y)

    if fill:
        ax.contourf(data, origin="lower", **kwargs)
    else:
        ax.contour(data, origin="lower", **kwargs)
    # ax.contour(data, **kwargs)
    # ax.set_xlim(xlim)
    # ax.set_ylim(ylim)


# def lazy_raster(imagename: str, **kwargs) -> None:
#     """
#     Rasterizes an image with matplotlib from a CASA style image file.
#     With this function, you can get an pretty image with a few lines of code.

#     Args:
#         imagename (str): CASA style image file.
#         **kwargs: Plot configuration keywords. Please see PlotConfig for more details.
#     """
#     imagename, img, config = prepare_image(imagename, **kwargs)
#     # plot
#     if not config.show or img.is_cube:
#         plt.ioff()

#     # Preparatoin for the cube
#     if img.is_cube and config.cbar == 'common':
#         if config.vmin is None:
#             config.vmin = np.nanmin(img.img)
#         if config.vmax is None:
#             config.vmax = np.nanmax(img.img)

#     # Specify the channel
#     if config.chan is None:
#         id_hz = 0
#     else:
#         id_hz = config.chan

#     fig, ax = plt.subplots(figsize=(8, 6))

#     # Draw the image
#     if img.is_cube:
#         im = ax.imshow(img.img[id_hz], cmap=config.cmap, aspect='equal', vmin=config.vmin, vmax=config.vmax, origin='lower')
#     else:
#         im = ax.imshow(img.img, cmap=config.cmap, aspect='equal', vmin=config.vmin, vmax=config.vmax, origin='lower')

#     # Draw the beam
#     draw_beam(ax=ax, img=img, color=config.cbeam)

#     # Set the title
#     if config.title is None:
#         config.title = imagename

#     # Set the axes options
#     set_axes_options(ax, config.title, img.axisname_x + f'[{unitDict[img.axis_unit_x]}]',
#                      img.axisname_y + f'[{unitDict[img.axis_unit_y]}]',
#                      *img.get_ticks(config.xtickspan, config.ytickspan, config.relative, config.ticksfmt))

#     # Set the colorbar
#     if config.cbarunit is None:
#         config.cbarunit = img.im_unit
#     set_cbar(im, config.cbarquantity, config.cbarunit, config.rescale, config.cbarfmt, ':.2f')
#     fig.tight_layout()

#     # Save the figure
#     savename = config.savename
#     if savename is not None:
#         if savename == '':
#             savename = imagename
#         if img.is_cube:
#             savename += f'-{id_hz}'
#         savename += '.png'
#         fig.savefig(savename, dpi=config.dpi)
#         print(f'Saved as "{savename}"')
#     if config.show:
#         plt.show()



from .Image import Image
from .io import load_fits, load_image
from .radial_profile import radial_profile
from .imshow import imshow, overlay_contour
from .imstat import imstat
from .radial_cut import radial_cut
from .detectpeak import detectpeak
from .azimuth_cut import azimuth_cut


import math
import os
import numpy as np

unitConvDict = {
    ('rad', 'rad'): 1,
    ('arcsec', 'arcsec'): 1,
    ('deg', 'deg'): 1,
    ('rad', 'arcsec'): (3600 * 180) / math.pi,
    ('arcsec', 'rad'): math.pi / (3600 * 180),
    ('deg', 'arcsec'): 3600,
    ('arcsec', 'deg'): 1 / 3600,
}

unitDict = {
    'rad': '',
    'arcsec': '"',
    'deg': '°',
}

siPrefixDict = {
    'peta': { 'symbol': 'P', 'base10': 1e-15 },
    'tera': { 'symbol': 'T', 'base10': 1e-12 },
    'giga': { 'symbol': 'G', 'base10': 1e-9 },
    'mega': { 'symbol': 'M', 'base10': 1e-6 },
    'kilo': { 'symbol': 'k', 'base10': 1e-3 },
    'hecto': { 'symbol': 'h', 'base10': 1e-2 },
    'deca': { 'symbol': 'da', 'base10': 1e-1 },
    'None': { 'symbol': '', 'base10': 1 },
    'deci': { 'symbol': 'd', 'base10': 1e1 },
    'centi': { 'symbol': 'c', 'base10': 1e2 },
    'milli': { 'symbol': 'm', 'base10': 1e3 },
    'micro': { 'symbol': 'μ', 'symbol_alt': 'u', 'base10': 1e6 },
    'nano': { 'symbol': 'n', 'base10': 1e9 },
    'pico': { 'symbol': 'p', 'base10': 1e12 },
}

def get_si_prefix_symbol(prefix: str) -> str:
    """
    Returns the SI prefix symbol for a given prefix string.

    Args:
        prefix (str): The SI prefix string.

    Returns:
        str: The symbol associated with the given prefix.
    """
    try :
        return siPrefixDict[prefix]['symbol']
    except KeyError:
        return siPrefixDict['None']['symbol']

def get_si_prefix_base10(prefix: str) -> float:
    """
    Returns the base 10 multiplier for a given SI prefix string.

    Args:
        prefix (str): The SI prefix string.

    Returns:
        float: The base 10 multiplier associated with the given prefix.
    """
    try :
        return siPrefixDict[prefix]['base10']
    except KeyError:
        return siPrefixDict['None']['base10']

def get_pret_dir_name(dir: str) -> str:
    """
    Returns the last component of a directory path without trailing slashes.

    Args:
        dir (str): The directory path.

    Returns:
        str: The last component of the directory path.
    """
    return os.path.split(dir.rstrip('/'))[1]

def downsample_data(data: np.ndarray, sample_size: int) -> np.ndarray:
    """
    Downsamples a 2D numpy array by averaging over blocks of size sample_size.

    Args:
        data (np.ndarray): The 2D numpy array to downsample.
        sample_size (int): The size of the blocks to average over.

    Returns:
        np.ndarray: The downsampled 2D numpy array.
    """
    if data.ndim != 2:
        raise ValueError("Input data must be a 2D numpy array.")
    if sample_size <= 0:
        raise ValueError("Sample size must be a positive integer.")
    
    height, width = data.shape
    
    # Crop the data to make sure dimensions are divisible by sample_size
    width_crop = width - (width % sample_size)
    height_crop = height - (height % sample_size)

    # Crop the data with keeping the center
    x_start = (width - width_crop) // 2
    y_start = (height - height_crop) // 2
    data = data[y_start:y_start + height_crop, x_start:x_start + width_crop]

    # New width and height
    width_new = width_crop // sample_size
    height_new = height_crop // sample_size

    return data.reshape(height_new, sample_size, width_new, sample_size).mean(axis=(1, 3))


from math import sqrt, ceil, cos, radians, sin
import numpy as np
from .Image import Image
from .utilities import downsample_data

def radial_profile(
    img: Image,
    azimuth: tuple | None = None,
    sample_size: int = 5,
    inc: float = 0.0,         # inclination in degrees (0 = face-on)
    PA: float = 0.0,          # position angle in degrees (east of north)
    stokes: int = 0,         # Stokes parameter index
    chan: int = 0,           # Channel index
) -> tuple:
    """
    Extract a radial profile from an image.

    Args:
        img (Image): The Image object.
        azimuth (tuple): The azimuth range angle in degrees. If None, all azimuth angles are considered.
        sample_size (int, optional): The number of samples to take along the radial line. Defaults to 5.
        inc (float, optional): The inclination angle in degrees. Defaults to 0.0.
        PA (float, optional): The position angle in degrees. Defaults to 0.0.
        stokes (int, optional): Stokes parameter index. Defaults to 0.
        chan (int, optional): Channel index. Defaults to 0.

    Returns:
        tuple: A tuple of three numpy arrays:
            - The radial distance from the center.
            - The mean intensity.
            - The standard deviation of the intensity.
    """
    if img.width is None or img.height is None:
        raise ValueError("Image width or height is None.")
    # Calculate the center of the image
    center_x = img.width // 2
    center_y = img.height // 2

    # Calculate the beam size
    if img.beam is None:
        raise ValueError("The image does not have a beam size.")
    if img.incr_x is None or img.incr_y is None:
        raise ValueError("Image increment x or y is None.")
    img.convert_axes_unit('arcsec')
    beam_maj = img.beam[0] / np.abs(img.incr_x)
    beam_min = img.beam[1] / np.abs(img.incr_y)

    # Determine the larger beam size (to define the sampling step and width)
    beam_size = max(beam_maj, beam_min)
    # sampling_size = ceil(beam_size * beam_factor)

    # Convert azimuth angle to radians
    # azimuth is measured from the north
    if azimuth is None:
        azimuth = (0, 359)
    azimuth = ((azimuth[0] + 90) % 360, (azimuth[1] + 90) % 360)

    def is_in_azimuth_range(angle, azimuth):
        if azimuth[0] < azimuth[1]:
            return azimuth[0] <= angle <= azimuth[1]
        else:
            return azimuth[0] <= angle or angle <= azimuth[1]

    # Initialize the line cut
    line_r = np.arange(0, min(center_x, center_y), sample_size, dtype=float)
    line_mean = []
    line_std = []
    sample = [[] for _ in range(len(line_r))]

    # Convert inclination and PA to radians
    inc_rad = radians(inc)
    PA_rad = radians(PA)

    # extract the 2D data
    # get the dimensions of the image
    data: np.ndarray = img.get_two_dim_data(stokes, chan)

    # Downsample the data
    data = downsample_data(data, sample_size)
    center_x_new = data.shape[1] // 2
    center_y_new = data.shape[0] // 2

    for i in range(len(data)):
        for j in range(len(data[0])):
            # Shift to new center
            dx = j - center_x_new
            dy = i - center_y_new

            # Rotate by -PA
            x_rot = dx * cos(-PA_rad) - dy * sin(-PA_rad)
            y_rot = dx * sin(-PA_rad) + dy * cos(-PA_rad)

            # Deproject y
            y_deproj = y_rot / cos(inc_rad)

            # Deprojected radius
            r = sqrt(x_rot**2 + y_deproj**2)

            rad = np.degrees(np.arctan2(y_deproj, x_rot) % (2 * np.pi))
            if is_in_azimuth_range(rad, azimuth):
                idx = int(r)
                if idx >= len(line_r):
                    continue
                sample[idx].append(data[i, j])

    for i, s in enumerate(sample):
        # print(f'Line {i}: {np.max(s)}')
        if not s:
            line_mean.append(0)
            line_std.append(0)
        else:
            line_mean.append(np.nanmean(s))
            line_std.append(np.nanstd(s))

    return np.array(line_r) * abs(img.incr_x), np.array(line_mean), np.array(line_std)



from math import radians, cos, sin, ceil, sqrt
import numpy as np
from .Image import Image
from .utilities import downsample_data


def radial_cut(
    img: Image, azimuth: float, beam_factor: float = 0.5, stokes: int = 0, chan: int = 0
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Radial line cut of the image along the specified azimuth angle from the center.

    Args:
        img (Image): The Image object.
        azimuth (float): The azimuth angle in degrees.
        beam_factor (float, optional): Sampling size based on the beam size. Defaults to 0.5.
        stokes (int, optional): Stokes parameter index. Defaults to 0.
        chan (int, optional): Channel index. Defaults to 0.

    Returns:
        tuple: A tuple of three numpy arrays:
            - The radial distance from the center.
            - The mean intensity.
            - The standard deviation of the intensity.
    """
    # Calculate the center of the image
    if img.width is None or img.height is None:
        raise ValueError("Image width or height is None.")
    center_x = img.width // 2
    center_y = img.height // 2

    # Calculate the beam size
    if img.beam is None:
        raise ValueError("The image does not have a beam size.")
    if img.incr_x is None or img.incr_y is None:
        raise ValueError("Image increment x or y is None.")
    img.convert_axes_unit("arcsec")
    beam_x = img.beam[0] / np.abs(img.incr_x)
    beam_y = img.beam[1] / np.abs(img.incr_y)

    # Determine the larger beam size (to define the sampling step and width)
    beam_size = max(beam_x, beam_y)
    sampling_size = ceil(beam_size * beam_factor)

    # Convert azimuth angle to radians
    # azimuth is measured from the north
    azimuth_rad = radians(azimuth - 270)

    data: np.ndarray = img.get_two_dim_data(stokes=stokes, chan=chan)
    # data = downsample_data(data, sampling_size)

    # Initialize the line cut
    line_r = []
    line_mean = []
    line_std = []

    # Define the unit direction vector for the given azimuth
    direction_x = cos(azimuth_rad)
    direction_y = sin(azimuth_rad)

    # Loop over points along the line
    step = 0
    search_range = int(sqrt(sampling_size**2 + sampling_size**2) / 2 + 1)
    while True:
        # Calculate the current position
        x = int(center_x + step * direction_x)
        y = int(center_y + step * direction_y)
        rect_center = np.array([x, y])

        if x < 0 or x >= img.width or y < 0 or y >= img.height:
            break

        line_r.append(step * np.abs(img.incr_x))

        v_para = np.array([direction_x, direction_y]) * sampling_size * 0.5
        v_perp = np.array([-direction_y, direction_x]) * beam_size * 0.5
        verts = [
            rect_center + v_para + v_perp,
            rect_center + v_para - v_perp,
            rect_center - v_para - v_perp,
            rect_center - v_para + v_perp,
        ]
        edges = [
            verts[1] - verts[0],
            verts[2] - verts[1],
            verts[3] - verts[2],
            verts[0] - verts[3],
        ]

        def cross(a, b):
            return a[0] * b[1] - a[1] * b[0]

        # Append the pixel value to the line
        sample = []
        # if step == 0: 0 else -search_range
        for i in range(-search_range, search_range + 1):
            for j in range(-search_range, search_range + 1):
                px = x + i
                py = y + j
                inside = True
                for k in range(4):
                    edge = edges[k]
                    vp = np.array([px, py]) - verts[k]
                    if cross(edge, vp) > 0:
                        inside = False
                        break
                if inside and 0 <= px < img.width and 0 <= py < img.height:
                    sample.append(data[py, px])

        if sample:
            line_mean.append(np.mean(sample))
            line_std.append(np.std(sample))
        else:
            line_mean.append(None)
            line_std.append(None)

        step += sampling_size

    return np.array(line_r), np.array(line_mean), np.array(line_std)
